const jetpack = require('fs-jetpack')
const moment = require('moment');
const powertools = require('node-powertools');
const path = require('path')
const os = require('os');
const { get, set, lowerFirst } = require('lodash');
const uuid = require('uuid');
const Manager = require('web-manager');
const AccountResolver = new (require('web-manager/lib/account.js'))({
  utilities: {
    get: get,
    set: set,
  },
  dom: function () {},
});

const platformName = process.platform === 'darwin'
  ? 'mac'
  : (process.platform === 'win32' ? 'windows' : 'linux')

let shutdownHandled = false;
let devlopmentRMDownload = false;
let copyFromVolumeTimeout;
let restartManagerAttempts = 0;

async function Main(_options) {
  const self = _options.self;
  const options = _options.options;

  self.performance.mark('manager_initialize_main_start');

  const { ipcMain, app } = self.libraries.electron;
  const environment = self.isDevelopment ? 'development' : 'production'

  if (self.isDevelopment) {
    console.log(`\n\n\n\n\nLaunching ${app.getName()}... app=${app.getVersion()}, electron=${process.versions.electron}, node=${process.versions.node}`);
  }

  // Set options
  options.hideInitially = typeof options.hideInitially === 'undefined' ? false : options.hideInitially;
  options.openAtLogin = typeof options.openAtLogin === 'undefined' ? true : options.openAtLogin;
  options.setProtocolHandler = typeof options.setProtocolHandler === 'undefined' ? true : options.setProtocolHandler;
  options.hideIfOpenedAtLogin = typeof options.hideIfOpenedAtLogin === 'undefined' ? true : options.hideIfOpenedAtLogin;
  options.autoUpdateInterval = typeof options.autoUpdateInterval === 'undefined' ? 60000 * 60 * 12 : options.autoUpdateInterval;
  options.autoRestartInterval = typeof options.autoRestartInterval === 'undefined' ? 30 : options.autoRestartInterval;
  options.singleInstance = typeof options.singleInstance === 'undefined' ? true : options.singleInstance;
  options.registerRestartManager = typeof options.registerRestartManager === 'undefined' ? true : options.registerRestartManager;
  options.setupTrayMenu = typeof options.setupTrayMenu === 'undefined' ? true : options.setupTrayMenu;
  options.setupAppMenu = typeof options.setupAppMenu === 'undefined' ? true : options.setupAppMenu;

  // Set dev options
  options._downloadRestartManager = typeof options._downloadRestartManager === 'undefined' ? false : options._downloadRestartManager;
  options._openDevelopmentRestartManager = typeof options._openDevelopmentRestartManager === 'undefined' ? false : options._openDevelopmentRestartManager;

  self.initOptions = options;

  // properties
  self.needsToBeShown = true;

  self.performance.mark('manager_initialize_main_determineVisibility');

  await _determineVisibility(self)

  self.performance.mark('manager_initialize_main_initializeRemote');

  self.libraries.remote = self.libraries.remote || require(path.join(app.getAppPath(), 'node_modules', '@electron/remote/main'));
  self.libraries.remote.initialize();

  self._handlers = {
    onDeepLink: function () {},
    onDeepLinkFilter: function () {return false},
    _onDeepLink: function () {
      function _internallyHandled(url) {
        try {
          url = new URL(url);
          if (url.protocol === `${self.options.app.id}:` && url.host === 'electron-manager') {
            const command = url.searchParams.get('command') || '';
            const payload = JSON.parse(url.searchParams.get('payload') || '{}');
            if (command === 'user:authenticate') {
              self.sendToRegisteredRenderers(command, payload);
              self.window().show('main')
            }
            return true;
          }
        } catch (e) {
          return false;
        }
      }

      if (self.deeplinkingUrl) {
        if (Array.isArray(self.deeplinkingUrl)) {
          for (var i = 0, l = self.deeplinkingUrl.length; i < l; i++) {
            const s = self.deeplinkingUrl[i];
            if (s.startsWith(`${self.options.app.id}://`) || self._handlers.onDeepLinkFilter(s)) {
              self.log('onDeepLink()', s)
              if (_internallyHandled(s)) {
                continue
              }
              self._handlers.onDeepLink(s)
            }
          }
        } else {
          if (_internallyHandled(self.deeplinkingUrl)) {
            return
          }
          self._handlers.onDeepLink(self.deeplinkingUrl)
        }
      }
    },
    onSecondInstance: function () {},
    _onSecondInstance: function () {
      if (self.secondInstanceParameters) {
        self._handlers.onSecondInstance(self.secondInstanceParameters)
      }
    },
    onLog: function () {},
    _onLogIsSet: false,
    _loggerQueue: [],

  };

  if (self.isDevelopment) {
    app.setPath('userData', `${app.getPath('userData')} (Development)`);
  }

  self.performance.mark('manager_initialize_main_openStorage');

  self._openStorage();
  const unverifiedConfig = self.storage.electronManager.get('data.config.data', {})
  let newData = self.storage.electronManager.get('data.current', {})
  self.storage.electronManager.set('data.previous', newData);

  // console.log('----unverifiedConfig', unverifiedConfig);
  // Init new sentry if desired
  self.performance.mark('manager_initialize_main_setupSentry');
  if (unverifiedConfig.sentry && unverifiedConfig.sentry.dsn) {
    self._sentryConfig.dsn = unverifiedConfig.sentry.dsn;
    self.libraries.sentry.init(self._sentryConfig);
  }

  self.performance.mark('manager_initialize_main_setupUA');
  app.userAgentFallback = app.userAgentFallback
    .replace(`Electron/${process.versions.electron}`, '')
    .replace(`${app.name}/${app.getVersion()}`, '')
    .replace(/\s\s+/g, ' ');

  function _shutdownHandler(code) {
    if (shutdownHandled) {
      return
    }
    if (code === 0) {
      self.storage.electronManager.set('data.current.usage.shutDownSafely', true);
      self.setupRestartManager('unregister');
    }
    const usage = self.usage().calculate({round: false, number: true})
    self.storage.electronManager.set('data.current.usage.opens', usage.total.opens + 1);
    self.storage.electronManager.set('data.current.usage.hours', usage.total.hours + usage.session.hours);

    if (self.isDevelopment) {
      console.log('Saving usage...', usage);
      app.removeAsDefaultProtocolClient(self.options.app.id);
    }
    shutdownHandled = true;
  }

  self.performance.mark('manager_initialize_main_setListeners1');
  app.on('quit', (event, code) => {
    _shutdownHandler(code)
  })

  if (process.platform === 'darwin') {
    process.on('exit', () => {
      _shutdownHandler(0)
    });
  }

  function getDeviceId() {
    return new Promise(function(resolve, reject) {
      const macaddress = require('macaddress');
      macaddress.one(async function (err, mac) {
        if (err) {
          console.error('Failed to get mac address', err);
          return resolve(uuid.v4());
        }
        return resolve(mac);
      })
    });
  }

  function getDeviceIp() {
    return new Promise(function(resolve, reject) {
      return resolve('0.0.0.0')
    });
  }

  // Initialize Global
  self.performance.mark('manager_initialize_main_initializeGlobal');
  newData = {
    meta: {
      startTime: new Date().toISOString(),
      // systemColorPreference: nativeTheme.shouldUseDarkColors ? 'dark' : 'light',
      // theme: nativeTheme.shouldUseDarkColors ? 'dark' : 'light',
      environment: environment,
      // edition: packageJSON.edition || 'standard',
      version: app.getVersion(),
      ip: await getDeviceIp(),
      country: 'ZZ',
      deviceId: await getDeviceId(),
      userAgent: app.userAgentFallback,
      os: {
        username: os.userInfo().username,
        platform: os.platform(),
        name: platformName,
        version: process.getSystemVersion(),
        locale: app.getLocale(),
        arch: process.arch,
      },
      indicator: {
        soft: '',
        strict: '',
      },
    },
    usage: {
      // opens: get(newData, 'usage.opens', 0) + 1,
      opens: get(newData, 'usage.opens', 1),
      hours: get(newData, 'usage.hours', 0),
      shutDownSafely: false,
      wasRelaunched: false,
    },
    paths: {
      mainDir: __dirname,
      userData: app.getPath('userData'),
      appData: app.getPath('appData'),
      downloads: app.getPath('downloads'),
      // module: app.getPath('module'),
      temp: app.getPath('temp'),
      // extensions: 'SET BELOW',
      // dependencies: 'SET BELOW',
      // resources: 'SET BELOW'
    },
    uuid: get(newData, 'uuid', null) || uuid.v4(),
    user: AccountResolver._resolveAccount(),
  };
  if (self.isDevelopment) {
    newData.user.roles.developer = true;
    newData.argv = require('yargs').argv
    console.log('Arguments', newData.argv);
    // newData.meta.startTime = "2022-06-14T06:52:43.844Z" // @@@DELETE
    // console.log('Usage', self.usage().calculate({round: true, number: false}));

  }
  self.storage.electronManager.set('data.current', newData);
  // console.log('---A', new Date().toISOString());
  // console.log('---B', newData);

  // setInterval(function () {
  //   console.log(self.storage.electronManager.get('data.current.meta'));
  // }, 1000);

  // main
  // setInterval(function () {
  //   self.storage.electronManager.set('data.current.uuid', Math.random())
  //   console.log('====uuid', self.storage.electronManager.get('data.current.uuid', null));
  // }, 1000);
  // setInterval(function () {
  //   console.log('====uuid', self.storage.electronManager.get('data.current.uuid', null));
  // }, 1000);

  self.performance.mark('manager_initialize_main_setProperties');
  self.windows = [];
  self.allowQuit = false;
  self.isQuitting = false;
  self.deeplinkingUrl = null;
  self.secondInstanceParameters = null;

  self._globalListenersWCs = self._globalListenersWCs || [];
  self._registeredRenderers = self._registeredRenderers || [];
  self._registeredRenderersIds = self._registeredRenderersIds || {}; // To make sure duplicate IDs are not registered
  self._updateTimeout;

  self._alertedConfig = false;


  self.onLog = (fn) => {
    self._handlers.onLog = fn;
    self._handlers._onLogIsSet = true;
  }

  if (self.options.log) {
    self._handlers._loggerQueue = [];
    self.log = function () {
      console.log(...self._addLogData(arguments));
      if (self._handlers._onLogIsSet) {
        if (self._handlers._loggerQueue.length > 0) {
          for (var i = 0; i < self._handlers._loggerQueue.length; i++) {
            self._handlers.onLog(...self._handlers._loggerQueue[i])
          }
          self._handlers._loggerQueue = [];
        }
        self._handlers.onLog(...arguments)
      } else {
        self._handlers._loggerQueue.push(self._addLogData(arguments))
      }
    }
  } else {
    if (self.isDevelopment) {
      self.log = function () {
        console.log('Called Manager.log() but options.log=false...', ...arguments);
      }
    } else {
      self.log = function () {}
    }
  }

  // if (self.options.log) {
  //   self._loggerQueue = [];
  //   self._log = function () {
  //     console.log(...self._addLogData(arguments));
  //     if (self.log) {
  //       if (self._loggerQueue.length > 0) {
  //         for (var i = 0; i < self._loggerQueue.length; i++) {
  //           self.log(...self._loggerQueue[i])
  //         }
  //         self._loggerQueue = [];
  //       }
  //       self.log(...arguments)
  //     } else {
  //       self._loggerQueue.push(self._addLogData(arguments))
  //     }
  //   }
  //   // self.log = function () {
  //   //   self._log(...arguments)
  //   //   // console.log('----HERE');
  //   //   // self._loggerQueue.push(self._addLogData(arguments))
  //   // }
  // } else {
  //   if (self.isDevelopment) {
  //     // self.log = function () {
  //     //   console.log('Called .log() but options.log=false', ...arguments);
  //     // }
  //     self._log = function () {
  //       console.log('Called ._log() but options.log=false', ...arguments);
  //     }
  //   } else {
  //     // self.log = function () {}
  //     self._log = function () {}
  //   }
  // }

  // if (self.options.log) {
  //   self._loggerQueue = [];
  //   self._log = function () {
  //     console.log(...self._addLogData(arguments));
  //     if (self.log) {
  //       if (self._loggerQueue.length > 0) {
  //         for (var i = 0; i < self._loggerQueue.length; i++) {
  //           self.log(...self._addLogData(self._loggerQueue[i]))
  //         }
  //         self._loggerQueue = [];
  //       }
  //       self.log(...self._addLogData(arguments))
  //     } else {
  //       self._loggerQueue.push(arguments)
  //     }
  //   }
  // } else {
  //   self._log = function () {}
  // }

  self.performance.mark('manager_initialize_main_initializeRendererStorage');

  self.libraries.electronStore.initRenderer();


  self.performance.mark('manager_initialize_main_instanceLock');

  const gotTheLock = app.requestSingleInstanceLock()

  if (!gotTheLock && !process.mas) {
    if (options.singleInstance) {
      self.allowQuit = true;
      self.isQuitting = true;
      app.exit();
    } else {
      // handle second instance deep link here
    }
  } else {
    app.on('second-instance', (event, commandLine, workingDirectory) => {
      if (process.platform !== 'darwin') {
        self.deeplinkingUrl = commandLine;
      }
      self._handlers._onDeepLink()

      self.secondInstanceParameters = {event: event, commandLine: commandLine, workingDirectory: workingDirectory}
      self._handlers._onSecondInstance()
    })

    // app.on('ready', function () {
    //   createMainWindow();
    // })
  }

  self.performance.mark('manager_initialize_main_setListeners2');
  app.on('will-finish-launching', function () {
    // Protocol handler for osx
    app.on('open-url', function(event, url) {
      event.preventDefault()
      self.deeplinkingUrl = url;
      self._handlers._onDeepLink()
    })
    app.on('open-file', function(event, path) {
      event.preventDefault()
      self.deeplinkingUrl = path;
      self._handlers._onDeepLink()
    })
  })

  if (process.platform !== 'darwin') {
    self.deeplinkingUrl = process.argv;
    self._handlers._onDeepLink()
  }

  // @@@ options.hideIfOpenedAtLogin USED TO BE HERE

  ipcMain.handle('electron-manager-message', async (event, message) => {
    message = message || {};
    message.payload = message.payload || {};

    // console.log('=====HERE', message);
    self.log('electron-manager-message', message)

    // if (message.command === 'global:get') {
    //   // console.log('=====GET 2', message.payload.path, message.payload.value);
    //   return self.global().get(message.payload.path, message.payload.value)
    // } else if (message.command === 'global:set') {
    //   return self.global().set(message.payload.path, message.payload.value)
    // } else if (message.command === 'global:register') {
    //   const senderWc = self.libraries.electron.webContents.fromId(get(event, 'sender.id', -1))
    //   if (senderWc) {
    //     self._globalListenersWCs = self._globalListenersWCs.concat(senderWc)
    //   }
    //   return self._global;
    // } else
    if (message.command === 'renderer:register') {
      const senderId = get(event, 'sender.id', -1);
      const senderWc = self.libraries.electron.webContents.fromId(senderId)
      if (senderWc && !self._registeredRenderersIds[senderId]) {
        self._registeredRenderers = self._registeredRenderers.concat(senderWc)
        self._registeredRenderersIds[senderId] = true;
      }
    } else if (message.command === 'renderer:show') {
      const win = self.window().get(event.sender.id);
      if (!win) {return}
      if (win.main) {
        // self.log(`[Performance] showMainWindow ${new Date().toISOString()}`);
        self.performance.mark('manager_initialize_main_showMainWindow');

        self.app().wasOpenedAtLogin()
        .then(wasOpenedAtLogin => {
          // If its an auto-show from preloader, only do it if we want to
          if (message.payload.source === 'preloader' && win.showDuringPreload && !wasOpenedAtLogin) {
            win.browserWindow.show();
          }
        })
      } else if (win.showDuringPreload) {
        win.browserWindow.show();
      }
    } else if (message.command === 'renderer:is-ready') {
      const win = self.window().get(event.sender.id);
      if (win) {
        win.ready = true;
        win._internal.handlers.onReady()
      }
    } else if (message.command === 'renderer:get-initial-data') {
      return {
        appPath: app.getAppPath(),
        isDevelopment: self.isDevelopment,
      }
    } else if (message.command === 'activity:last-action') {
      clearTimeout(self._updateTimeout);
      self._updateTimeout = setTimeout(function () {
        self.log('In-activity triggered auto-update')
        self.libraries.updater.update();
      }, options.autoUpdateInterval);
    } else if (message.command === 'main:process-deep-link') {
      self.deeplinkingUrl = message.payload.url;
      self._handlers._onDeepLink()
    } else if (message.command === 'main:rebuild-menus') {
      if (self.libraries.trayMenu.initialized) {
        self.libraries.trayMenu.init();
      }
      if (self.libraries.appMenu.initialized) {
        self.libraries.appMenu.init();
      }
    } else if (message.command === 'app:get-app-path') {
      return app.getAppPath();
    } else if (message.command === 'app:relaunch') {
      self.relaunch(message.payload);
    } else if (message.command === 'app:quit') {
      self.quit(message.payload)
    } else if (message.command === 'updater:update') {
      self.libraries.updater.update(message.payload)
    } else if (message.command === 'special:alert-config') {
      if (!self._alertedConfig) {
        self.libraries.electron.dialog.showMessageBox(undefined, {
          message: `${self.options.app.name} is using config: \n\n ${JSON.stringify(message.payload, null, 4)}`,
          type: 'info'
        })
        self._alertedConfig = true;
      }
    }
  })

  self.performance.mark('manager_initialize_main_setOpenAtLogin');
  if (options.openAtLogin) {
    await self.app().setLoginItemSettings().catch(e => console.error);
  }
  if (!options.openAtLogin || self.isDevelopment) {
    await self.app().setLoginItemSettings({openAtLogin: false}).catch(e => console.error);
  }  

  self.performance.mark('manager_initialize_main_setProtocolHandler');
  if (options.setProtocolHandler) {
    await self.app().setAsDefaultProtocolClient(self.options.app.id).catch(e => console.error);
    if (self.isDevelopment) {
      app.removeAsDefaultProtocolClient(self.options.app.id);
    }    
  }

  // const mainLibrary = new (require('./libraries/analytics.js'))(self, appPath);
  // self.libraries.analytics = await self.libraries.analytics.init(get(self.options, 'config.analytics', {}));

  self.performance.mark('manager_initialize_main_setupHelpers');
  self.sendToRegisteredRenderers = function (command, payload) {
    self._registeredRenderers
    .forEach((wc, i) => {
      wc.send('electron-manager-message', {
        command: command,
        payload: payload,
      })
    });
  }

  self.relaunch = function (options) {
    options = options || {}
    if (options.force) {
      self.allowQuit = true;
    }

    self.storage.electronManager.set('data.current.usage.wasRelaunched', true);

    try {
      if (self.libraries.updater.status.code === 'downloaded') {
        return self.libraries.updater.library.quitAndInstall();
      } else {
        return app.quit(app.relaunch());
      }
    } catch (e) {
      return app.quit(app.relaunch());
    }
  }

  self.quit = function (options) {
    options = options || {}
    if (options.force) {
      self.allowQuit = true;
    }
    return app.quit();
  }

  self.window = function () {
    return {
      create: function (options) {
        const win = {};
        options = options || {};
        options.preferences = options.preferences || {};
        options.preferences.show = typeof options.preferences.show === 'undefined' ? false : options.preferences.show

        win.main = typeof options.main === 'undefined' ? false : options.main;
        win.ready = typeof options.ready === 'undefined' ? false : options.ready;
        win.allowWindowOpens = typeof options.allowWindowOpens === 'undefined' ? !!win.main : options.allowWindowOpens;
        win.handleCloseEvents = typeof options.handleCloseEvents === 'undefined' ? !!win.main : options.handleCloseEvents;
        win.handleNeedsToBeShown = typeof options.handleNeedsToBeShown === 'undefined' ? !!win.main : options.handleNeedsToBeShown;
        win.enableRemoteModule = typeof options.enableRemoteModule === 'undefined' ? !!win.main : options.enableRemoteModule;
        win.preventDevTools = typeof options.preventDevTools === 'undefined' ? !!win.main : options.preventDevTools;

        let failedWindowTimeout;

        if (win.main) {
          self.performance.mark('manager_initialize_main_createBrowserWindow');
          failedWindowTimeout = setTimeout(function () {
            self.libraries.updater.update();
          }, 1000 * 30);
        }

        if (typeof options.showDuringPreload === 'undefined') {
          win.showDuringPreload = win.main ? !self.initOptions.hideInitially : true;
        } else {
          win.showDuringPreload = options.showDuringPreload;
        }

        win.createLoggerFunction = typeof options.createLoggerFunction === 'undefined' ? !!win.main : options.createLoggerFunction;

        win._internal = {
          lockedUrl: '',
          handlers: {
            onReady: () => {},
          }
        }

        // Handlers        
        win.onReady = function (fn) {
          win._internal.handlers.onReady = function () {
            self.performance.mark('manager_initialize_main_browserWindowReady');
            clearTimeout(failedWindowTimeout);
            fn()
          };
        }
        win.browserWindow = new self.libraries.electron.BrowserWindow(options.preferences)
        if (win.browserWindow && win.browserWindow.webContents) {
          win.id = win.browserWindow.webContents.id || -1;
        }

        if (!win.allowWindowOpens) {
          win.browserWindow.webContents.setWindowOpenHandler(() => {
            return {action: 'deny'}
          })
        }
        if (win.handleCloseEvents) {
          win.browserWindow.on('close', function (event) {
            self.log('win.close event', win.id);
            // console.log('----event', event);
            if (!self.allowQuit && !self.isQuitting) {
              win.browserWindow.hide();
              return event.preventDefault();
            }
          });

          win.browserWindow.on('closed', function (event) {
            self.log('win.closed event', win.id);
            // console.log('----event', event);
            win.browserWindow = null;
            self.libraries.electron.app.quit();
          })
        }
        if (win.handleNeedsToBeShown && self.needsToBeShown) {
          win.browserWindow.once('show', async (event) => {
            self.log('win.show event', win.id);
            if (process.platform === 'darwin') {
              app.show();
              app.dock.show();
            }
          });
        }

        if (win.preventDevTools) {
          win.browserWindow.webContents.on('devtools-opened', function (event) {
            if (!self.storage.electronManager.get('data.current.user.roles.developer', false)) {
              win.browserWindow.webContents.closeDevTools()
              setTimeout(function () {
                if (win && win.browserWindow && win.browserWindow.webContents) {
                  win.browserWindow.webContents.closeDevTools()
                }
              }, 10);
              return event.preventDefault();
            }
          });
        }

        if (win.main) {
          self.libraries.electron.app.on('activate', function () {
            self.log('win.activate event', win.id);
            // On macOS it's common to re-create a window in the app when the
            // dock icon is clicked and there are no other windows open.
            if (win.browserWindow) {
              win.browserWindow.show();
            }
          })
        }

        if (win.enableRemoteModule) {
          self.libraries.remote.enable(win.browserWindow.webContents);
        }
        
        if (self.isDevelopment) {
          win.browserWindow.webContents.openDevTools({ mode: 'undocked', activate: false });
        }

        if (win.createLoggerFunction) {
          win.browserWindow.webContents.once('dom-ready', async function (event) {
            await powertools.poll(function () {
              return win && !!win.ready;
            }, {interval: 1000, timeout: self.isDevelopment ? 10000 : 200000})
            .catch(e => {
              console.error('Failed to create logger functions because win.ready never happened');
            })
            self.onLog(function () {
              // console.log('Logger', ...arguments);
              if (win && win.browserWindow && win.browserWindow.webContents) {
                win.browserWindow.webContents.send('electron-manager-message', {command: 'console:log', payload: [...arguments]})
              }
            })
            self.log('Setup automatic logger')
          });
        }

        self.windows.push(win);
        return win;
      },
      get: function (id) {
        if (id) {
          if (typeof id === 'number') {
            return self.windows.filter(w => w.id === id)[0];
          } else {
            return self.windows.filter(w => w[id])[0];
          }
        } else {
          return self.windows;
        }
      },
      send: function (id, name, message, options) {
        return new Promise(async function(resolve, reject) {
          options = options || {};
          options.wait = typeof options.wait === 'undefined' ? true : options.wait;

          let win = self.window().get(id);

          if (options.wait) {
            try {
              await powertools.poll(function () {
                win = self.window().get(id);
                return win && !!win.ready;
              }, {interval: 1000, timeout: 30000})
              .catch(e => reject(e))
            } catch (e) {
              return reject(e);
            }
          }
          // console.log('------SEND', win);

          if (win && win.browserWindow && win.browserWindow.webContents) {
            win.browserWindow.webContents.send(name, message)
            return resolve();
          } else {
            return reject(new Error(`Window with id ${id} does not exist`));
          }
        });
      },
      navigate: function (id, url, options) {
        const win = self.window().get(id);
        if (win && win.browserWindow && win.browserWindow.webContents) {
          options.lock = typeof options.lock === 'undefined' ? !!win.main : options.lock;

          if (win.main) {
            self.performance.mark('manager_initialize_main_browserWindowNavigate');
          }

          if (options.lock && !win._internal.lockedUrl) {
            win._internal.lockedUrl = url;
            win.browserWindow.webContents.on('will-navigate', function (event, url) {
              // console.log('NAV', url, win._internal.lockedUrl);
              if (url !== win._internal.lockedUrl) {
                console.error('Prevented navigation because this window is locked:', url, win._internal.lockedUrl);
                event.preventDefault();
              }
            });
          }

          let isFile = false;
          try {
            isFile = new URL(url).protocol === 'file:';
          } catch (e) {
            isFile = true;
          }

          if (isFile) {
            return win.browserWindow.loadFile(url, options)
          } else {
            return win.browserWindow.loadURL(url, options)
          }

        } else {
          return new Promise(function(resolve, reject) {
            reject(new Error(`Window ${id} does not exist`))
          });
        }

      },
      show: function (id) {
        const win = self.window().get(id);
        if (win && win.browserWindow) {
          if (win.browserWindow.isMinimized()) {
            win.browserWindow.restore();
          }
          if (!win.browserWindow.isVisible() || !win.browserWindow.isFocused()) {
            win.browserWindow.show();
          }
        } else {

        }
      },
      hide: function (id) {
        const win = self.window().get(id);
        if (win && win.browserWindow) {
          if (!win.browserWindow.isMinimized()) {
            win.browserWindow.minimize();
          }
        } else {

        }
      },
      toggle: function (id) {
        const win = self.window().get(id);
        if (win && win.browserWindow) {
          if (win.browserWindow.isMinimized() || !win.browserWindow.isVisible() || !win.browserWindow.isFocused()) {
            self.window().show(id);
          } else {
            self.window().hide(id);
          }
        } else {

        }
      },
    }
  };


  self.setupRestartManager = async function (command) {
    self.log('setupRestartManager()', command);
    if (self.options.app.id === 'restart-manager') {
      return
    }
    if (restartManagerAttempts >= 3) {
      return
    }
    restartManagerAttempts++;
    const restartURL = new URL('restart-manager://message')
    const restartPayload = {
      name: app.getName(),
      id: self.options.app.id,
      path: app.getPath('exe'),
      environment: environment,
      // appPath: app.getAppPath(),
      // argv: process.argv,
      // execPath: process.execPath,
      // portableExeDirectory: process.env.PORTABLE_EXECUTABLE_DIR,
    }
    restartURL.searchParams.set('command', command);
    restartURL.searchParams.set('payload', JSON.stringify(restartPayload));
    // console.log('Opening restart-manager', restartURL.toString());
    // console.log('----restartPayload', restartPayload);

    function handleRMError() {
      self.log('handleRMError()');
      const downloadURL = {
        // mac: 'https://github.com/somiibo/download-server/releases/download/installer/Somiibo.dmg',
        // mac: 'https://github.com/itw-creative-works/restart-manager-download-server/releases/download/installer/Restart-Manager.zip',
        mac: 'https://github.com/itw-creative-works/restart-manager-download-server/releases/download/installer/Restart-Manager.dmg',

        // windows: 'https://github.com/somiibo/download-server/releases/download/installer/Somiibo-Setup.exe',
        windows: 'https://github.com/itw-creative-works/restart-manager-download-server/releases/download/installer/Restart-Manager-Setup.exe',

        // linux: 'https://github.com/somiibo/download-server/releases/download/installer/Somiibo-amd64.deb',
        linux: 'https://github.com/itw-creative-works/restart-manager-download-server/releases/download/installer/restart-manager_amd64.deb',
      }[platformName];
      const localPathFilename = path.resolve(app.getPath('appData'), 'Restart Manager', 'resources', downloadURL.split('/').pop())
      const localPathArray = localPathFilename.split('/');
      localPathArray.pop()
      const localPath = path.join('/', ...localPathArray);
      const localPathAppname = path.join(localPath, 'Restart Manager.app');

      function executeInstalled() {
        self.log('executeInstalled()');
        return new Promise(async function(resolve, reject) {
          const extractZip = require('./zip.js');
          const util = require('util')
          const execute = util.promisify(require('child_process').exec);

          if (process.platform === 'darwin') {
            // jetpack.remove(path.resolve(localPath, 'Restart Manager.app'));
            // await extractZip(localPathFilename, localPath)
            // extractZip('/Users/ianwiedenman/Library/Application Support/Restart Manager/resources/download.zip', '/Users/ianwiedenman/Library/Application Support/Restart Manager/resources')
            await execute(`open "${localPathFilename}"`);

            async function copyFromVolume() {
              try {
                clearTimeout(copyFromVolumeTimeout)
                const volumes = jetpack.list('/Volumes').sort().reverse();
                const found = volumes.find(v => v.includes('Restart Manager'));

                if (found) {
                  // console.log('----1');
                  process.noAsar = true;
                  if (jetpack.exists(localPathAppname)) {
                    // console.log('----2');
                    jetpack.remove(localPathAppname);
                  }
                  // console.log('----3');
                  // jetpack.remove(localPathAppname);
                  // console.log('----4');
                  jetpack.copy(path.join('/Volumes', found, 'Restart Manager.app'), localPathAppname)
                  // console.log('----5');
                  await execute(`open "${localPathAppname}"`);
                  ejectDisks()
                  return resolve();
                } else {
                  copyFromVolumeTimeout = setTimeout(function () {
                    copyFromVolume();
                  }, 3000);
                }
              } catch (e) {
                console.error(e);
                return resolve();
              }
              process.noAsar = undefined
            }

            async function ejectDisks() {
              await execute(`diskutil list`)
              .then(async result => {
                const split = result.stdout.split('\n');
                for (var i = 0; i < split.length; i++) {
                  const line = split[i];
                  if (line.includes('Restart Manager')) {
                    const split2 = line.split(' ');
                    for (var i2 = 0; i2 < split2.length; i2++) {
                      const item2 = split2[i2];
                      if (item2.includes('disk')) {
                        console.log(`Ejecting disk ${item2}`);
                        await execute(`diskutil eject ${item2}`)
                        .catch(e => console.error('Failed to eject disks', e))
                      }
                    }
                  }
                }
              })
              .catch(e => console.error('Failed to list disks', e))
            }

            await ejectDisks()

            copyFromVolume()
          } else if (process.platform === 'win32') {
            await execute(`"${localPathFilename}"`);
            return resolve();
          } else {
            // TODO: INSTALL WITH SNAP?
            // if (true) {
            //   sudo snap install restart-manager
            // }
            await execute(`sudo apt install "${localPathFilename}"`).catch(e => {console.error(e)})
            await execute(`restart-manager`).catch(e => {console.error(e)})
            return resolve();
          }
        });
      }

      // executeInstalled();
      // return

      downloadItem(self, downloadURL, localPathFilename)
      .catch(e => {
        self.log(`Download restart-manager failed: ${e}`);
      })
      .finally(async () => {

        executeInstalled()
        .then(() => {
          setTimeout(function () {
            self.setupRestartManager(command)
          }, 5000);
        })

      })
    }

    // handleRMError();
    // return

    function openRM() {
      return new Promise(async function(resolve, reject) {
        const restartManagerHandler = await self.app().getApplicationNameForProtocol('restart-manager')
        self.log(`openRM(), handler=${restartManagerHandler}`);
        if (restartManagerHandler === 'Electron' && !options._openDevelopmentRestartManager) {
          console.log('Skipping restart manager because in development')
          return resolve();
        } else if (restartManagerHandler) {
          self.libraries.electron.shell.openExternal(restartURL.toString())
          .then(() => {
            return resolve();
          })
          .catch((e) => {
            return reject(e);
          })
        } else {
          return reject(new Error('restart-manager handler does not exist'))
        }
      });
    }

    openRM()
      .then(() => {
        if (self.isDevelopment && !devlopmentRMDownload && options._downloadRestartManager) {
          devlopmentRMDownload = true;
          handleRMError();
        }
      })
      .catch(e => {
        self.log(`Error managing restart-manager 1: ${e}`);

        setTimeout(function () {
          openRM()
            .catch(e => {
              self.log(`Error managing restart-manager 2: ${e}`);
              handleRMError();
            })
        }, self.isDevelopment ? 1000 : 60000);
      })
  }

  // self.analytics(app.getAppPath());
  self.performance.mark('manager_initialize_main_setupAnalytics');
  self.analytics();

  self.performance.mark('manager_initialize_main_setupMenus');
  self.libraries.trayMenu = new (require('./tray-menu.js'))(self);
  if (options.setupTrayMenu) {
    self.libraries.trayMenu.init();
  }

  self.libraries.appMenu = new (require('./app-menu.js'))(self);
  if (options.setupAppMenu) {
    self.libraries.appMenu.init();
  }

  self.libraries.updater = new (require('./updater.js'))(self);
  self.libraries.updater.init();

  self.performance.mark('manager_initialize_main_registerRestartManager');
  if (options.registerRestartManager) {
    app.whenReady().then(() => {
      self.performance.mark('manager_initialize_main_appIsReady');
      setTimeout(function () {
        self.setupRestartManager('register');
      }, self.isDevelopment ? 3000 : 15000);
    })
  }

  self.performance.mark('manager_initialize_main_setupAutoRestart');

  // Every day check if app has been open for more than X days. if so, restart it
  if (options.autoRestartInterval > 0) {
    setInterval(function () {
      const days = moment().diff(moment(newData.meta.startTime), 'days', true)
      console.log(`Daily open check: ${days} days`);
      if (days >= options.autoRestartInterval) {
        self.relaunch({force: true})
      }
    }, 1000 * 60 * 60 * 24);
  }

  // setTimeout(function () {
  //   self.log('Initialized Electron Manager');
  // }, 1);

  self.performance.mark('manager_initialize_main_setupDebugger');

  if (unverifiedConfig.debug && unverifiedConfig.debug.throwMainError) {
    const debug = new (require('./debug.js'))
    setTimeout(function () {
      debug.throw(unverifiedConfig.debug.throwMainError, `Config error #${unverifiedConfig.debug.errorMessage || '0'} - ${new Date().toLocaleString()}`)
    }, 10);
  }

  // DONE WITH Main
  // self.log(`[Performance] init (main): ${self.started.toISOString()}`);
  self.performance.mark('manager_initialize_main_end');
  return
}

function downloadItem(self, downloadURL, localPathFilename) {
  return new Promise(function(resolve, reject) {
    const { BrowserWindow } = self.libraries.electron;
    const win = new BrowserWindow({ show: false });
    self.log('downloadItem()', downloadURL, localPathFilename);

    try {
      jetpack.remove(localPathFilename);
    } catch (e) {

    }

    win.webContents.session.on('will-download', (event, item, webContents) => {
      // Set the save path, making Electron not to prompt a save dialog.
      // item.setSavePath('/tmp/save.pdf')
      item.setSavePath(localPathFilename)

      item.on('updated', (event, state) => {
        if (state === 'interrupted') {
          console.log('Download is interrupted but can be resumed')
        } else if (state === 'progressing') {
          if (item.isPaused()) {
            console.log('Download is paused')
          } else {
            console.log(`Received bytes: ${(item.getReceivedBytes() / item.getTotalBytes()) * 100}`)
          }
        }
      })
      item.once('done', (event, state) => {
        if (state === 'completed') {
          console.log('Download successfully', localPathFilename);
          return resolve()
        } else {
          console.log(`Download failed: ${state}`)
          return reject()
        }
        win.close();
      })
    })

    win.loadURL(downloadURL)
    .catch(e => {})
  });
}

function whoIsMyDaddy() {
  try {
    throw new Error('a');
  } catch (e) {
    // matches this function, the caller and the parent
    const allMatches = e.stack.match(/(\w+)@|at (\w+) \(/g);
    // match parent function name
    const parentMatches = allMatches[2].match(/(\w+)@|at (\w+) \(/);
    // return only name
    return parentMatches[1] || parentMatches[2];
  }
}

function _determineVisibility(self) {
  return new Promise(async function(resolve, reject) {
    const { app } = self.libraries.electron;

    if (self.initOptions.hideIfOpenedAtLogin) {
      const wasOpenedAtLogin = await self.app().wasOpenedAtLogin();
      if (wasOpenedAtLogin) {
        return resolve()
      } 
    }

    if (self.initOptions.hideInitially) {
      return resolve();
    }

    if (platformName === 'mac') {
      app.show();
      app.dock.show();
      self.needsToBeShown = false;
    } 

    return resolve();
  });
}

module.exports = Main;
